//////////////////////////////////////////////////////////////////////
//
//  ASSERT.CH
//
//  Copyright:
//      Alaska Software, (c) 1998-2023. All rights reserved.         
//  
//  Contents:
//      #command directives for the pseudo function ASSERT()
//   
//  Remarks:
//      The pseudo function ASSERT() appears in the source code if the
//      #define constant NDEBUG is not defined. ASSERT() is used to
//      check a logical condition during program development for debugging
//      purposes. If an assertion (logical condition) fails during
//      program execution an error message is displayed.
//   
//      There are two modes for ASSERT() which depend on the #define
//      constant _ASSERT_QUIT:
//   
//      If _ASSERT_QUIT is defined the program terminates once an
//      assertion fails (default).
//   
//      If _ASSERT_QUIT is not defined an alert() box is displayed
//      with the choices "Ignore" and "Quit".
//   
//      All ASSERT() pseudo functions are removed from source code
//      when the constant NDEBUG is defined or if a file is compiled
//      using the compiler switch /dNDEBUG .
//   
//  Syntax:
//      ASSERT( <lCondition> [, <cMessage>] )
//   
//        Examples:
//          ASSERT( VALTYPE(nValue)=="N", "Wrong data type" )
//          ASSERT( .NOT. EMPTY(aArray) , "Array has no elements" )
//   
//////////////////////////////////////////////////////////////////////


// Assert.ch is not included
#ifndef  _ASSERT_CH       

// NDEBUG is not defined: create code for ASSERT()
#ifndef  NDEBUG           

// Comment this line if you want to choose between {"Ignore","Quit"}
// when an assertion fails
   #define  _ASSERT_QUIT  


// The ASSERT() pseudo function terminates a program when an assertion fails
   #ifdef  _ASSERT_QUIT   

      #command  ASSERT( <condition> [, <message>] ) ;
            =>  IF .NOT. (<condition>) ;;
                   QOut( "Error in:", ProcName(0) ) ;;
                   QOut( "    Line:", ProcLine(0) ) ;;
                   QOut( #<message> )  ;;
                   QUIT ;;
                ENDIF

// ASSERT() with {"Ignore","Quit"}
   #else                  

      #command  ASSERT( <condition> [, <message>] ) ;
            =>  IF .NOT. (<condition>) ;;
                   Eval( {|pn,pl,ms,n| n := Alert( ;
                           "Error in: "+ pn +";"+ ;
                           "    Line: "+ Ltrim(Str(pl)) +";"+ ms, ;
                           {"Ignore","Quit"}), ;
                           IIF(n==1,NIL,_Quit()) ;
                         }, ProcName(0) , ProcLine(0) , #<message>  ;
                       ) ;;
                ENDIF

   #endif       // #define _ASSERT_QUIT

// Ignore ASSERT() -> NDEBUG is defined!
#else                     

   #command  ASSERT( <condition> [, <message>] )  =>

#endif          // #define NDEBUG


// Assert.ch is included
#define  _ASSERT_CH       

#endif          // #ifndef _ASSERT_CH

