#!/usr/bin/env python3
from email import encoders
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import mimetypes
import os
import re
import smtplib
import imaplib
import datetime
import time
import configparser
import sys
print(sys.version)

class Log:
    """ Support class used to write log
    """
    def __init__(self, fname):
        self.fname = fname

    def email_log(self, email):
        d = datetime.datetime.now()
        f = open(self.fname,'a')
        f.write('%s\t%s\t%s\t%s\n' % (d.strftime('%Y-%m-%d %H:%M'),email._mail['smtp'],email._to[0],email._subject))
        f.close()

    def error_log(self, e, err_type='error', throw=True):
        d = datetime.datetime.now()
        f = open(self.fname,'a')
        f.write('%s\t%s %s\n' % (d.strftime('%Y-%m-%d %H:%M'),err_type,str(e)))
        f.close()
        if throw:
            raise e

    def txt_log(self, s, throw=True):
        d = datetime.datetime.now()
        f = open(self.fname,'a')
        f.write('%s\t%s\n' % (d.strftime('%Y-%m-%d %H:%M'),s))
        f.close()
        if throw:
            raise Exception(s)


class Email:
    """ This class handles the creation and sending of email messages via SMTP.  
    This class also handles attachments and can send HTML messages.  The code 
    comes from various places around the net and from my own brain.
    """
    def __init__(self, log, mail, srv):
        """ Create a new empty email message object.
        @param logfile: Name of logfile to store info about traffic.
        @type logfile: String
        """
        self._textBody = None
        self._htmlBody = None
        self._notify = None
        self._log = log
        self._mail = mail
        self._srv = srv
        self._subject = ""
        self._reEmail = re.compile("^([\\w \\._]+\\<[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\\>|[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?)$")
        self.clearRecipients()
        self.clearAttachments()
    
    def send(self, server, sender = None):
        """ Send the email message represented by this object.
        @param server: The SMTP server used to send the mail.
        @type server: smtplib.SMTP Object
        """
        # Validate message
        if self._textBody is None and self._htmlBody is None:
            raise Exception("Error! Must specify at least one body type (HTML or Text)")
        if len(self._to) == 0:
            raise Exception("Must specify at least one recipient")
        
        # Create the message part
        if self._textBody is not None and self._htmlBody is None:
            msg = MIMEText(self._textBody, "plain")
        elif self._textBody is None and self._htmlBody is not None:
            msg = MIMEText(self._htmlBody, "html")
        else:
            msg = MIMEMultipart("alternative")
            msg.attach(MIMEText(self._textBody, "plain"))
            msg.attach(MIMEText(self._htmlBody, "html"))
        # Add attachments, if any
        if len(self._attach) != 0:
            tmpmsg = msg
            msg = MIMEMultipart()
            msg.attach(tmpmsg)
        for fname,attachname in self._attach:
            if not os.path.exists(fname):
                print("File '%s' does not exist.  Not attaching to email." % fname)
                continue
            if not os.path.isfile(fname):
                print("Attachment '%s' is not a file.  Not attaching to email." % fname)
                continue
            # Guess at encoding type
            ctype, encoding = mimetypes.guess_type(fname)
            if ctype == 'text/xml':
                ctype = 'application/xml'
            if not ctype:
                ctype = ''
            ctype += '/'
            if ctype is None or encoding is not None:
                # No guess could be made so use a binary type.
                ctype = 'application/octet-stream'
            maintype, subtype = ctype.split('/', 1)
            if maintype == 'text':
                fp = open(fname)
                attach = MIMEText(fp.read(), _subtype=subtype)
                fp.close()
            elif maintype == 'image':
                fp = open(fname, 'rb')
                attach = MIMEImage(fp.read(), _subtype=subtype)
                fp.close()
            elif maintype == 'audio':
                fp = open(fname, 'rb')
                attach = MIMEAudio(fp.read(), _subtype=subtype)
                fp.close()
            else:
                fp = open(fname, 'rb')
                attach = MIMEBase(maintype, subtype)
                attach.set_payload(fp.read())
                fp.close()
                # Encode the payload using Base64
                encoders.encode_base64(attach)
            # Set the filename parameter
            if attachname is None:
                filename = os.path.basename(fname)
            else:
                filename = attachname
            attach.add_header('Content-Disposition', 'attachment', filename=filename)
            msg.attach(attach)
        # Some header stuff
        msg['Subject'] = self._subject
        if sender:
            def_sender = self._from
            self._from = sender
        msg['From'] = self._from
        msg['To'] = ", ".join([i.strip() for i in self._to if i])
        if self._notify:
            if self._notify == 'me':
                msg['Disposition-Notification-To'] = self._from
            else:
                msg['Disposition-Notification-To'] = self._notify
        ## msg['X-Mailer'] = 'mailsend.py by http://compusof.it'  # forse questo manda in spam
        msg.preamble = "You need a MIME enabled mail reader to see this message"
        # Send message
        msg = msg.as_string()
        server.sendmail(self._from, self._to, msg)
        self._log.email_log(self)
        if sender:
            self._from = def_sender
        return msg

    def notify(self, address='me'):
        self._notify = address
    
    def setSubject(self, subject):
        """ Set the subject of the email message.
        """
        self._subject = subject
    
    def setFrom(self, address):
        """ Set the email sender.
        """
        if not self.validateEmailAddress(address):
            self._log.txt_log("Invalid sender email '%s'" % address)
        self._from = address
    
    def clearRecipients(self):
        """ Remove all currently defined recipients for the email message.
        """
        self._to = []
    
    def addRecipient(self, address):
        """ Add a new recipient to the email message.
        """
        if not self.validateEmailAddress(address):
            self._log.txt_log("Invalid recipient email '%s'" % address)
        self._to.append(address)
    
    def setTextBody(self, body):
        """ Set the plain text body of the email message.
        """
        self._textBody = body
    
    def setHtmlBody(self, body):
        """ Set the HTML portion of the email message.
        """
        self._htmlBody = body
    
    def clearAttachments(self):
        """ Remove all file attachments.
        """
        self._attach = []
    
    def addAttachment(self, fname, attachname=None):
        """ Add a file attachment to this email message.

        @param fname: The full path and file name of the file
                      to attach.
        @type fname: String
        @param attachname: This will be the name of the file in
                           the email message if set.  If not set
                           then the filename will be taken from
                           the fname parameter above.
        @type attachname: String
        """
        if fname is None:
            return
        self._attach.append( (fname, attachname) )
    
    def validateEmailAddress(self, address):
        """  Validate the specified email address.
        
        @return: True if valid, False otherwise
        @rtype: Boolean
        """
        if self._reEmail.search(address.lower()) is None:
            return False
        return True


def sendmail(mail):
    """ send an email, the mail parameter is a dict build reading a file
    with txt_decode(name), 
    """
    ini = configparser.ConfigParser(interpolation=None)
    ini.read(mail['ini'])
    log = Log(ini.get('MAIL','logfile'))
    try:
        time_out = ini.getint('MAIL','timeout')
    except Exception as e:
        time_out = 20

    # get server parameters from mail['smtp'] section or from default server
    if not 'smtp' in mail:
        mail['smtp'] = ini.get('MAIL','server')
    srv = dict(ini.items(mail['smtp']))

    # open the mail server with declared cryptography
    try:
        if not 'port' in srv:
            srv['port'] = 25
        if 'security' in srv and srv['security'] == 'SSL':
            server = smtplib.SMTP_SSL(srv['host'],srv['port'],timeout=time_out)
        else:
            server = smtplib.SMTP(srv['host'],srv['port'],timeout=time_out)
        if 'security' in srv and srv['security'] == 'TLS':
            server.starttls()
        if 'username' in srv:
            server.login(srv['username'],srv['password'])
    except Exception as e:
        log.error_log(e,'server error')

    # initialize the email object
    m = Email(log, mail, srv)
    # first search sender in mail['smtp'] then in srv and alt last in 
    # [MAIL] section of .ini file 
    sender = mail.get('sender', '')
    if not sender:
        sender = srv.get('sender','')
    if not sender:
        sender = ini.get('MAIL','sender') 
    m.setFrom(sender)

    for i in mail['to'].split('\n'):
        try:
            m.addRecipient(ini.get('MAIL','debug'))
        except Exception as e:
            m.addRecipient(i)

    m.setSubject(mail['subject'])
    try:
        if mail['body']:
            m.setTextBody(mail['body'])
    except:
        pass
    try:
        if mail['html']:
            m.setHtmlBody(mail['html'])
    except:
        pass
    try:
        for i in mail['attach'].split('\n'):
            m.addAttachment(i)
    except:
        pass
    try:
        for i in mail['filename'].split('\n'):
            m.addAttachment(i)
    except:
        pass
    try:
        if mail['notify'].upper().find('YES') != -1:
            m.notify()
    except:
        pass
    # *** send the mail ***
    try:
        msg = m.send(server)
    except Exception as e:
        log.error_log(e,'email error')

    # *** forward to the sender ***
    try:
        if mail['forward'].upper().find('YES') != -1:
            m.clearRecipients()
            m.addRecipient(mail['sender'])
            m.setSubject('fwd: '+mail['subject'])
            try:
                m.send(server)
            except Exception as e:
                log.error_log(e,'email error')
    except:
        pass

    server.quit()
    return msg

def imap_sent(mail, msg):
    """
    send a copy of email to the sent folder of imap account of the sender
    """
    ini = configparser.ConfigParser()
    ini.read(mail['ini'])
    sender = ini.get('MAIL','sender')
    if not ini.has_section('IMAP_'+sender):
        return
    t = ini.items('IMAP_'+sender)
    srv = {}
    for i in t:
        srv[i[0]] = i[1]
    if not 'host' in srv:
        return
    if not 'security' in srv:
        srv['security'] = ''
    if not 'port' in srv:
        srv['port'] = 25
    if srv['security'] == 'SSL':
        m = imaplib.IMAP4_SSL(srv['host'],srv['port'])
    else:
        m = imaplib.IMAP4(srv['host'],srv['port'])
    m.login(srv['username'],srv['password'])
    if sys.version_info[0] > 2:
        msg = str.encode(msg)
    m.append(srv['sent_folder'], '\\Seen', imaplib.Time2Internaldate(time.time()),msg)
    m.logout()
    log = Log(ini.get('MAIL','logfile'))
    log.txt_log("Copy to sent %s (%s)" % (sender, srv['sent_folder']), throw=False)

def txt_decode(name):
    """
    Read a text file and build a dict with sections and text.
    each section begin with '[' and ends with ']'
    """
    f = open(name,'r')
    l = f.readlines()
    l = [i.strip() for i in l]
    d = {}    
    for i in l:
        ## if i[0] == ';':
        ##     continue
        if i and i[0] == '[' and i[-1] == ']':
            key = i[1:-1]
            d[key] = ''
            continue
        if d[key] != '':
            d[key] += '\n'
        d[key] += i
    return d
    
if __name__ == "__main__":
    if len(sys.argv) > 1:
        name = sys.argv[1]
    else:
        print("File name missed!")
        quit()
    d = txt_decode(name)
    msg = sendmail(d)
    imap_sent(d, msg)
